//  Mazeing
//  routines associated with drawing, painting or printing the made

#include "stdafx.h"
#include "mazeing.h"

#include "mainwnd.h"

#include "globals.h"


// draw a maze on a dc
// dc must have been scaled (ie 8 units to one maze cell) and the correct pen selected
// used for OnPaint, printing and creating the meta file for copying to the clipboard
void DrawMaze(CDC *dc)
{
	register int x, y, e ;

    if(ValidMaze) {
    
		DrawingDone = 0 ;

		// draw the horizontal lines
		for(x = 0 ; x < xd ; x++) {
			DrawingDone++ ;
			for(y = 0 ; y < yd ; y++)
				if(CELL(x, y) & LFT) {
					for(e = y ; e < yd && (CELL(x, e) & LFT) ; e++)  ;
					DrawWall(dc, x, y, 0, 0, (e-y)<<3, FALSE) ;
					y = e ;
					if(AbortPrint) return ;
				}
			}
					
		// draw the verticle lines
		for(y = 0 ; y < yd ; y++) {
			DrawingDone++ ; 
			for(x = 0 ; x < xd ; x++)
				if(CELL(x, y) & TOP) {
					for(e = x ; e < xd && (CELL(e, y) & TOP) ; e++) ;
					DrawWall(dc, x, y, 0, 0, (e-x)<<3, TRUE) ;
					x = e ;
					if(AbortPrint) return ;
				}
			}
					
		// draw the entrance and exit lines
		for(x = 0 ; x < xd ; x++)
			for(y = 0 ; y < yd ; y++) {
				if(CELL(x, y) & ENTRY) {
					DrawWall(dc, x, y, 0, 0, 1, TRUE) ;
					DrawWall(dc, x, y, 1, -1, 2, FALSE) ;
					DrawWall(dc, x, y, 7, -1, 2, FALSE) ;
					DrawWall(dc, x, y, 7, 0, 1, TRUE) ;
					}
				}
		}
}





// Draw a wall of the maze, called by only DrawMaze
void DrawWall(CDC *dc, int x, int y, int dx, int dy, int len, int horiz)
{
	dc->MoveTo((x<<3) + dx, (y<<3) + dy) ;
    if(horiz)
    	dc->LineTo((x<<3) + dx + len, (y<<3) + dy) ;
	else
    	dc->LineTo((x<<3) + dx, (y<<3) + dy + len) ;
}





// Draw the solution
// dc must have been scaled (ie 8 units to one maze cell) and the correct pen selected
// used for OnPaint, printing and creating the meta file for copying to the clipboard
void DrawSolution(CDC *dc)
{
	int e ;
		
	dc->MoveTo((PathX[0]<<3) + 4, (PathY[0]<<3) - 4) ;
	dc->LineTo((PathX[0]<<3) + 4, (PathY[0]<<3) + 4) ;
	for(e = 1 ; e <= PathLength ; e++)
		dc->LineTo((PathX[e]<<3) + 4, (PathY[e]<<3) + 4) ;
	dc->LineTo((PathX[e-1]<<3) + 4, (PathY[e-1]<<3) + 12) ;
}	





// Draw ONE line segment of the track, uses i as an index into TPath?[]
// dc must have been scaled (ie 8 units to one maze cell) and the correct pen selected
// used for OnPaint, printing and creating the meta file for copying to the clipboard
void DrawTrackLine(CDC *dc, int i)
{
	dc->MoveTo((TPathX[i-1]<<3) + 4, (TPathY[i-1]<<3) + 4) ;
	dc->LineTo((TPathX[i]<<3) + 4, (TPathY[i]<<3) + 4) ;
}




	
// abort proceedure.
// called by the windows printing routined to check if printing should be aborted
HWND HPrintCancelDlg = NULL;

BOOL FAR PASCAL __export AbortProc (HDC HPrnDC, int Code)
{
	MSG Msg ;
	while (!AbortPrint && ::PeekMessage (&Msg, NULL, 0, 0, PM_REMOVE))
		if (!HPrintCancelDlg || !::IsDialogMessage (HPrintCancelDlg, &Msg)) {
			::TranslateMessage (&Msg) ;
			::DispatchMessage (&Msg);
			}                        
	return (!AbortPrint) ;
}


                                                          
                                                          
// main printing routine called by OnPrint and OnAutoPrint
// this gets the printing dc, sets up the abort box, calls DrawMaze to print, etc
BOOL DoPrintMaze(int Copies, BOOL PrtSolution, BOOL PrtTracks, CString Title)
{	
	int err ;
	CDC dc ;
	char *msg = NULL ;
	LPDEVNAMES lpDevNames = NULL ;
	LPDEVMODE  lpDevMode = NULL ; 
	HDC hDC = NULL ;
	int r, z, t, b ;		// margin sizes in pixels
	CString str = "grg" ;
	CSize TextSize ;
	
    // get a handle to the printer DC.  This code stolen from the MFC as I could
    // not get the standard CPrintDialog to get it for me!!!
	lpDevNames = (LPDEVNAMES)::GlobalLock(pPrintDialog->m_pd.hDevNames);
	lpDevMode = (pPrintDialog->m_pd.hDevMode != NULL) ? 
						(LPDEVMODE)::GlobalLock(pPrintDialog->m_pd.hDevMode) : NULL;

	if (lpDevNames == NULL) { err = -101 ; goto PrintError ; }
	
	hDC = ::CreateDC((LPCSTR)lpDevNames + lpDevNames->wDriverOffset,
					  (LPCSTR)lpDevNames + lpDevNames->wDeviceOffset,
					  (LPCSTR)lpDevNames + lpDevNames->wOutputOffset,
					  lpDevMode);

	// Global unlocks not needed in protect mode
	::GlobalUnlock(pPrintDialog->m_pd.hDevNames);
	::GlobalUnlock(pPrintDialog->m_pd.hDevMode);

	if(hDC == NULL || !dc.Attach(hDC)) { err = -102 ; goto PrintError ; }

	// get the page margins	
	if((r = GetMarginSize(&dc, MRgt)) == -1) { str = MRgt ; MRgt = "1.5 cm" ;}
	if((z = GetMarginSize(&dc, MLft)) == -1) { str = MLft ; MLft = "1.5 cm" ;}
	if((t = GetMarginSize(&dc, MTop)) == -1) { str = MTop ; MTop = "1.5 cm" ;}
	if((b = GetMarginSize(&dc, MBot)) == -1) { str = MBot ; MBot = "1.5 cm" ;}
	if(str != "grg") {
		NotifyMsg( "Error in page margin : %s\r\nValue is too large or not expressed correctly\r\nUse a number followed by in for inches, cm or mm for metric\r\n\nValue is reset to the default  1.5 cm", str) ;
		WriteProfileMargins() ;
		AutoPrintMazes = 0 ;
		return FALSE ;
		}

	// scale the page for drawing
	dc.SetMapMode(MM_ISOTROPIC) ;
	dc.SetWindowExt((xd-1)<<3, (yd-1)<<3) ;
	dc.SetViewportExt(dc.GetDeviceCaps(HORZRES) - r - z, dc.GetDeviceCaps(VERTRES) - t - b) ;
	dc.SetViewportOrg(z, t) ;
    
    /*// set the number of copies using the printer driver if possible
	if(Copies > 1) {
		int ActualCopies, One = 1 ;
		err = dc.Escape(SETCOPYCOUNT, sizeof(int), (LPSTR)&Copies, (LPSTR)&ActualCopies) ;
		if(!err || ActualCopies != Copies) // printer or print driver does not support multiple copies
			dc.Escape(SETCOPYCOUNT, sizeof(int), (LPSTR)&One, (LPSTR)&ActualCopies) ;
		else
			Copies = 1 ;
		}
	 */
	// create and setup the abort print dialog box
	pMainWnd->EnableWindow(FALSE);
	if(!PrintCancelDlg.Create(IDD_PRINTCANCEL))
		{ err = -100 ; goto PrintError ; }
	PrintCancelDlg.m_PrinterName.SetWindowText(pPrintDialog->GetDeviceName() 
							+ CString(" on ") + pPrintDialog->GetPortName()) ;
	PrintCancelDlg.m_Msg.SetWindowText(Title) ;
	AbortPrint = 0;
	HPrintCancelDlg = PrintCancelDlg.m_hWnd;
	if((err = dc.SetAbortProc(AbortProc)) < 0) goto PrintError ;

	// start the printing process
	if((err = dc.StartDoc(CString("Mazeing:  ") + Title)) < 0) goto PrintError ;

	while(Copies--) {
		if((err = dc.StartPage()) < 0) goto PrintError ;
		
		// scale the page for drawing
		// this is a test to correct a bug reported by Tony Errington
		dc.SetMapMode(MM_ISOTROPIC) ;
		dc.SetWindowExt((xd-1)<<3, (yd-1)<<3) ;
		dc.SetViewportExt(dc.GetDeviceCaps(HORZRES) - r - z, dc.GetDeviceCaps(VERTRES) - t - b) ;
		dc.SetViewportOrg(z, t) ;
	    
		TextSize = dc.GetTextExtent("M", 1) ;
		dc.TextOut(0, -TextSize.cy, CString("Mazeing V1.0   ") + Title) ;
	
		LockArrays() ;
	
	    DrawMaze(&dc) ;

		if(!AbortPrint && PrtTracks) {
			CPen *OldPen, TrackPen(PS_DOT, 1, 0) ;
			OldPen = dc.SelectObject(&TrackPen) ;
			for(int i = 1 ; i < TrackLength ; i++)
				DrawTrackLine(&dc, i) ;
			dc.SelectObject(OldPen) ;
			}
				
		if(!AbortPrint && PrtSolution) {
			CPen *OldPen, SolutionPen(PS_DASH, 1, 0) ;
			OldPen = dc.SelectObject(&SolutionPen) ;
			DrawSolution(&dc) ;
			dc.SelectObject(OldPen) ;
			}
			
		UnLockArrays() ;
		if(AbortPrint) {err = SP_USERABORT ; goto PrintError ;}
		
		if((err = dc.EndPage()) < 0) goto PrintError ;
		} // while copies to print

	if((err = dc.EndDoc()) < 0) goto PrintError ;

	// cleanup
	pMainWnd->EnableWindow(TRUE) ;
	PrintCancelDlg.DestroyWindow() ;
	return TRUE ;
	
 // abort routine, used for all errors
 // print a usefull message and exit false
 PrintError:
	switch(err) {
		case 0:
		case SP_USERABORT:
		case SP_APPABORT:	msg = NULL ; break ;
		case SP_OUTOFDISK:	msg = "Not enough disk space for printing" ; break ;
		case SP_OUTOFMEMORY:msg = "Not enough memory for printing" ; break ;
		case -100:			msg = "Cannot create print cancel dialog box" ; break ;
		case -101:			msg = "Cannot get printer name" ; break ;
		case -102:			msg = "Cannot create a printer device context" ; break ;
		default:			msg = "Unknown cause" ;  break ;
		}
	AutoPrintMazes = 0 ;
	pMainWnd->EnableWindow(TRUE) ;
	PrintCancelDlg.DestroyWindow() ;
	if((err & SP_NOTREPORTED) && msg != NULL) NotifyMsg("Printing error.\r\n%s.", msg) ; 
	return FALSE ;
}





// get a page margin given the string entered by the user
// returns margin size in pixels given the dc and a string
// the string can end with in, cm, mm to indicate the dimension
int GetMarginSize(CDC *dc, CString Str)
{
	int i ;
	long r, mm ;
	
	Str.MakeLower() ;

	// skip trailing spaces	
	for(i = Str.GetLength() -1 ; i > 0 ; i--)
		if(Str[i] != ' ' || Str[i] != '\t')
			break ;
			
	if(Str.Mid(i-1,2) == "mm")
		mm = atoi(Str) ;
	else if(Str.Mid(i-1,2) == "cm")
		mm = (int)(atof(Str) * 10.0) ;
	else if(Str.Mid(i-1,2) == "in")
		mm = (int)(atof(Str) * 25.4) ;
	else
		mm = -1 ;
		
	r = ((long)dc->GetDeviceCaps(HORZRES) * mm) / (long)dc->GetDeviceCaps(HORZSIZE) ;
	
	if(r >= 0 && r < (long)dc->GetDeviceCaps(HORZRES)/3L)
		return (int)r ;
	else
		return -1 ;
}





